#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# fake gdb command line interface window for the 
# SmartGDB interface.
# Author: Charles Dymond  <cdy@ittc.ukans.edu>
# command line entries can be typed into the window
# and are passed along to smartGDB, commands that
# change the state of other GUI windows are intercepted
# and executed by other GUI routines
# "But its a NICE wheel!"
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
global commandLine
set commandLine 1

global commandWindow
set commandWindow 0

global commandBuffer
set commandBuffer ""

global prompt
set prompt "(gdb) "

global lastCommand
set lastCommand ""

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# maximum number of commands 
# remembered by the command history
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
global MAX_NUM_COMMANDS
set MAX_NUM_COMMANDS 256

global commandHist
set commandHist ""

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# Note: the script prompt must be the same
# length as the standard prompt (including spaces)
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
global script_prompt
set script_prompt "----> "

global promptOffset
set promptOffset 6


proc tkerror { errmsg } {

    global commandWindow
    global commandBuffer
    insert_text .smartgdb.status "Error: $errmsg"
    if { $commandWindow == 1 } {
	set msg "invalid command, wrong planetary alignment, or other error (don't panic) "
	append msg \n
	set commandBuffer ""
	echo_output $msg
	echo_output $errmsg
    }
}


# ============================================================================
# Procedure: newputs
# Purpose: For the interface we need to rename puts and use one that
#	   just returns its value so we can capture it on the screen.
# ============================================================================
proc new_puts { a } {
    .gdb.txt insert insert $a
    new_line
}

 
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# displays output in the command window
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
proc echo_output { blurb } {
 
    global commandLine
    global promptOffset
    
    .gdb.txt insert insert $blurb
    new_line
    set inc [ expr [ llength [ split $blurb \n ] ] -2 ]
    set commandLine [expr $commandLine+$inc]
    .gdb.txt mark set insert $commandLine.0
    show_prompt
    .gdb.txt delete $commandLine.$promptOffset end

}


#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# Adds the latest command to the command history
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
proc update_cmnd { cmnd } {
    global commandBuffer
    global lastCommand
    global commandHist

    set lastCommand $commandBuffer

    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    # add command to command history list
    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    set commandHist [linsert $commandHist 0 $cmnd]
    set commandBuffer ""
}


#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# Processes a command line, commands bound to 
# interface buttons are invoked by executing
# the routines bound to those buttons.
# Other commands are first checked for Tcl
# completeness using SmartGDB's isComplete
# command. (Note: invalid commands are seen
# as complete (completely wrong); only
# partial Tcl commands/procedures are seen as
# incomplete)
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
proc process_line { line } {

    global commandWindow
    global commandBuffer
    global commandLine
    global script_prompt
    global lastCommand
    global insertFG
    global insertBG

    append commandBuffer $line

    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # if return is pressed and a new command has 
    # not been entered, repeat the last command
    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    if { $commandBuffer == "" } {
	set commandBuffer $lastCommand
    }

    switch  $commandBuffer {
	c -
	cont { do_cont; update_cmnd $commandBuffer }
	n -
	next { do_next; update_cmnd $commandBuffer }
	s -
	step { do_step; update_cmnd $commandBuffer }
	r -
	run { do_cont; update_cmnd $commandBuffer }
	cl -
	cls { 
	    .gdb.txt delete 1.0 end
	    set commandLine 1
	    show_prompt
	    update_cmnd $commandBuffer
	}
	exit { 
	    set commandWindow 0
	    destroy .gdb
	}
	default { 
	    set complete [ isComplete $commandBuffer ]
	    if { $complete == 1 } {
		set blurb [ eval $commandBuffer ]
		update_cmnd $commandBuffer
		if { $blurb != "" } {
		    echo_output $blurb
		} else {
		    show_prompt
		}
	    } else {
		append commandBuffer \n
		.gdb.txt insert insert \n
		.gdb.txt mark set insert $commandLine.0
		.gdb.txt insert insert $script_prompt 
		.gdb.txt mark set lf $commandLine.0
		.gdb.txt mark set ll $commandLine.end
		.gdb.txt tag add more lf ll
		.gdb.txt tag configure more -foreground $insertFG
		.gdb.txt tag configure more -background $insertBG
		.gdb.txt see $commandLine.0
	    }
	}		
    }
}


#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# insets a new line and updates the current
# line and insertion points
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
proc new_line { } {
    global commandLine
    .gdb.txt insert insert \n
    incr commandLine
    .gdb.txt mark set insert $commandLine.0
 }


#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# Returns the current command line minus
# the prompt
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
proc get_line { } {
    global commandLine
    global promptOffset

    .gdb.txt mark set first $commandLine.$promptOffset
    .gdb.txt mark set last $commandLine.end
    set line [.gdb.txt get first last]
    return $line
}


#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# displays the prompt
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
proc show_prompt { } {
    global commandLine
    global prompt
    global promptFG
    global promptBG
    .gdb.txt insert insert $prompt
    .gdb.txt mark set lf $commandLine.0
    .gdb.txt mark set ll $commandLine.end
    .gdb.txt tag add prmpt lf ll
    .gdb.txt tag configure prmpt -foreground $promptFG \
	-background $promptBG
    .gdb.txt see $commandLine.0
}


#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# Displays a command from the command history based on
# how many times the up (or down) arrow key has been 
# pressed since the last time the "Enter" key has been
# pressed
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
proc show_cmndhist { } {

    global commandLine
    global cmnd
    global promptOffset
    global commandHist
    global MAX_NUM_COMMANDS

    if { $commandHist != "" } { 
	set length [llength $commandHist]
	set command [expr $cmnd-1 ]

	#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
	# if you get to the end of the list, go back
	# to the begining
	#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
	if {$command > $length} {
	    set command 0
	    set cmnd 0
	}

	#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
	# if the maximum length of the command history
	# has been exceeded, delete the last item in
	# the history
	#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
	if { $length > $MAX_NUM_COMMANDS } {
	    set commandHist [lreplace $commandHist end end]
	}

	#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
	# get the selected command and display it
	# on the command line
	#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
	set command [lindex $commandHist $command] 
	.gdb.txt delete $commandLine.$promptOffset end
	.gdb.txt insert insert $command
    }
}


#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# MAIN ENTRY point
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
# creates a standard x-term sized window with
# "fake" SmartGDB command line interface 
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
proc gdb_window { } {

    global commandWindow
    global commandLine
    global promptOffset
    global prompt
    global cmnd
    global commandFG
    global commandBG
    global commandCursor

    set promptOffset [string length $prompt]

    toplevel .gdb
    wm title .gdb "gdb command window"
    wm geometry .gdb 485x315+15+15
    
    text .gdb.txt \
	-fg $commandFG -bg $commandBG -wrap none \
	-xscrollcommand ".gdb.xscroll set" \
	-yscrollcommand ".gdb.yscroll set" \
	-insertwidth 5 -insertbackground $commandCursor

    scrollbar .gdb.xscroll -orient horizontal \
	-command ".gdb.txt xview"
    scrollbar .gdb.yscroll -orient vertical \
	-command ".gdb.txt yview"

    grid columnconfigure .gdb 0 -weight 1
    grid columnconfigure .gdb 1 -weight 0
    grid rowconfigure .gdb 0 -weight 1
    grid rowconfigure .gdb 1 -weight 0
    
    grid .gdb.txt -column 0 -row 0 -sticky nsew
    grid .gdb.xscroll -column 0 -row 1 -sticky ew
    grid .gdb.yscroll -column 1 -row 0 -sticky ns

    .gdb.txt mark set insert 1.0
    focus .gdb.txt

    set commandWindow 1
    set commandLine 1
    show_prompt

    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+- 
    # rename puts so the "puts" used in scripts will 
    # display to the commmand window
    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+- 
    catch { rename puts old_puts }
    catch { rename new_puts puts } 

    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-   
    # grab the command line (except for the prompt),
    # process it for commands and then delete any 
    # extra characters that tend to insert themselves 
    # following the command line              
    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-   
    bind .gdb.txt <Return> {
	set line [get_line]
	new_line
	process_line $line
	.gdb.txt delete $commandLine.$promptOffset end
	set cmnd 0
	break
    }

    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-      
    # a click anywhere sets the cursor to the 
    # command line insertion point 
    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    bind .gdb.txt <Button-1> {
	.gdb.txt mark set insert $commandLine.$promptOffset
	break
   }

    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    # this code ensures that that the user can't
    # backspace over the prompt(s)
    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    bind .gdb.txt <KeyPress-BackSpace> {
	set ind [.gdb.txt index insert]
	set ind [ lindex [ split $ind . ] 1 ]
	if { $ind == $promptOffset } {
	    break }
    }

    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    # increments cmnd counter for command history
    # and displays the next command in the history
    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    bind .gdb.txt <Up> {
	incr cmnd
	show_cmndhist
	break
    }


    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    # decrements the counter for the command history
    # and displays the previous command in the history
    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    bind .gdb.txt <Down> {
	if { $cmnd > 0 } {
	    set cmnd [expr $cmnd-1]
	}
	show_cmndhist
	break
    }

    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    # left arrow cannot move cursor into the prompt(s)
    #+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
    bind .gdb.txt <Left> {
	set ind [.gdb.txt index insert]
	set ind [ lindex [ split $ind . ] 1 ]
	if { $ind == $promptOffset } {
	    break }
    }

    bind .gdb.txt <B1-Motion> {
	break
    }
}

